/*
Copyright 2021 nxa07283, NXP
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    (1) Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

    (2) Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.  
    
    (3)The name of the author may not be used to
    endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

#ifndef _NTAG5_H_
#define _NTAG5_H_

#include <stdint.h>
#include <stdbool.h>






/* ----- NTAG5 API ----- */

struct ntag5_hw_config {
	int i2c_adapter;
	int i2c_address;
	int gpio;
};

#define NTAG_TYPE_I2CPLUS (4)
#define NTAG_TYPE_NTAG5 (5)

#define NTAG_I2C_ADDR_AUTO (0xFF)
#define NTAG_I2C_ADDR_NTAG5 (0x54)

struct ntag {
	int type;
	int i2c_file;
	int i2c_address;
};

#define NTAG5_MODE_NORMAL (0)
#define NTAG5_MODE_MIRROR (1)

typedef struct ntag * ntag_t;

/* ntag5.c */
ntag_t ntag5_open(int type, struct ntag5_hw_config * hw_config);
int ntag5_close(ntag_t ntag);
int ntag5_nfc_enable(ntag_t ntag, bool enable);
int ntag5_nfc_is_enabled(ntag_t ntag);
int ntag5_nfc_is_present(ntag_t ntag);
int ntag5_reset(ntag_t ntag);
int ntag5_set_mode(ntag_t ntag, int mode);
int ntag5_pthru_receive(ntag_t ntag, uint8_t *buf, int cnt);
int ntag5_pthru_send(ntag_t ntag, uint8_t *buf, int cnt);


/* ntag5_driver.c */
int ntag5_reg_read(ntag_t ntag, uint32_t reg, uint16_t *val);
int ntag5_reg_write(ntag_t ntag, uint32_t reg, uint16_t mask, uint16_t val);
int ntag5_block_read(ntag_t ntag, int block, uint8_t *bytes, int len);
int ntag5_block_write(ntag_t ntag, int block, uint8_t *bytes, int len);
int ntag5_mem_read(ntag_t ntag, uint16_t addr, uint8_t *bytes, int len);
int ntag5_mem_write(ntag_t ntag, uint16_t addr, uint8_t *bytes, int len);
int ntag5_cfg_read(ntag_t ntag, uint16_t block, uint8_t *bytes);
int ntag5_cfg_write(ntag_t ntag, uint16_t block, uint8_t *bytes);



/* ----- Memories */

#define NTAG5_EEPROM_BLOCKS			(0x0200)
#define NTAG5_EEPROM_BLOCK_SIZE		(4)
#define NTAG5_EEPROM_SIZE			(NTAG5_EEPROM_BLOCKS*NTAG5_EEPROM_BLOCK_SIZE)
#define NTAG5_EEPROM_START			(0x0000)
#define NTAG5_EEPROM_END			(0x01FF)


#define NTAG5_SRAM_BLOCKS			(0x40)
#define NTAG5_SRAM_BLOCK_SIZE		(4)
#define NTAG5_SRAM_SIZE				(NTAG5_SRAM_BLOCKS*NTAG5_SRAM_BLOCK_SIZE)
#define NTAG5_SRAM_START			(0x2000)
#define NTAG5_SRAM_END				(0x203F)

#define NTAG5_CONFIG_BLOCKS			(0xA0)
#define NTAG5_CONFIG_BLOCK_SIZE		(4)
#define NTAG5_CONFIG_SIZE			(NTAG5_CONFIG_BLOCKS * NTAG5_CONFIG_BLOCK_SIZE)
#define NTAG5_CONFIG_START			(0x1000)
#define NTAG5_CONFIG_END			(0x109F)

#define NTAG5_EEPROM(addr)			((addr) + NTAG5_EEPROM_START * NTAG5_EEPROM_BLOCK_SIZE)
#define NTAG5_SRAM(addr)			((addr) + NTAG5_SRAM_START * NTAG5_SRAM_BLOCK_SIZE)



/* Packed register encoding:
 *    [31] always 1 
 * [30:16] bits address
 *  [15:8] bits offset
 *   [7:0] bits width 
 * 
 * Bit 31 prevents user from supplying a page address instead of NTAG5_REG()
 */
 
 
 /* ----- Registers -----  */

 
#define NTAG5_REG(addr, off, w) ((1<<31) | (((addr) & 0x7FFF) << 16) | (((off) & 0x3) << 8) | ((w) & 0x3))
#define NTAG5_REG_CHECK(r) (((r) & (1<<31)) ? 1 : 0)
#define NTAG5_REG_GET_ADDR(reg) (((reg) >> 16) & 0x7FFF)
#define NTAG5_REG_GET_OFF(reg) (((reg) >> 8) & 0x3)
#define NTAG5_REG_GET_W(reg) ((reg) & 0x3)
#define NTAG5_REG_GET_ADDR_H(reg) (((reg) >> 24) & 0x7F)
#define NTAG5_REG_GET_ADDR_L(reg) (((reg) >> 16) & 0xFF)

#define NTAG5_STATUS_REG NTAG5_REG(0x10A0, 0, 2)
#define NTAG5_STATUS0_REG NTAG5_REG(0x10A0, 0, 1)
#define NTAG5_STATUS1_REG NTAG5_REG(0x10A0, 1, 1)

#define NTAG5_CONFIG_REG NTAG5_REG(0x10A1, 0, 3)
#define NTAG5_CONFIG_0_REG NTAG5_REG(0x10A1, 0, 1)
#define NTAG5_CONFIG_1_REG NTAG5_REG(0x10A1, 1, 1)
#define NTAG5_CONFIG_2_REG NTAG5_REG(0x10A0, 2, 1)

#define NTAG5_SYNC_DATA_BLOCK_REG NTAG5_REG(0x12A2, 0, 2)

#define NTAG5_PWM_GPIO_CONFIG_REG NTAG5_REG(0x12A3, 0, 2)
#define NTAG5_PWM_GPIO_CONFIG_0_REG NTAG5_REG(0x12A3, 0, 1)
#define NTAG5_PWM_GPIO_CONFIG_1_REG NTAG5_REG(0x12A3, 1, 1)

#define NTAG5_PWM0_ON_OFF_REG NTAG5_REG(0x10A4, 0, 4)
#define NTAG5_PWM0_ON_REG NTAG5_REG(0x10A4, 0, 2)
#define NTAG5_PWM0_OFF_REG NTAG5_REG(0x10A4, 2, 2)

#define NTAG5_PWM1_ON_OFF_REG NTAG5_REG(0x10A5, 0, 4)
#define NTAG5_PWM1_ON_REG NTAG5_REG(0x10A5, 0, 2)
#define NTAG5_PWM1_OFF_REG NTAG5_REG(0x10A5, 2, 2)

#define NTAG5_WDT_CONFIG_REG NTAG5_REG(0x10A6, 0, 2)
#define NTAG5_WDT_EN_REG NTAG5_REG(0x10A6, 2, 1)

#define NTAG5_EH_CONFIG_REG NTAG5_REG(0x10A7, 0, 1)
#define NTAG5_ED_CONFIG_REG NTAG5_REG(0x10A8, 0, 1)

#define NTAG5_I2C_SLAVE_ADDR_REG NTAG5_REG(0x10A9, 0, 1)
#define NTAG5_I2C_SLAVE_CONFIG_REG NTAG5_REG(0x10A9, 1, 1)

#define NTAG5_RESET_GEN_REG NTAG5_REG(0x10AA, 0, 1)

#define NTAG5_ED_INTR_CLEAR_REG NTAG5_REG(0x10AB, 0, 1)

#define NTAG5_I2C_S_M_ADD_REG NTAG5_REG(0x10AC, 0, 1)
#define NTAG5_I2C_M_LEN_REG NTAG5_REG(0x10AC, 1, 1)
#define NTAG5_ALM_STATUS_REG NTAG5_REG(0x10AC, 2, 1)

#define NTAG5_I2C_M_STATUS_REG NTAG5_REG(0x10AD, 0, 1)

#define NTAG5_STATUS1_VCC_BOOT_OK		(1 << 7)
#define NTAG5_STATUS1_NFC_BOOT_OK		(1 << 6)
#define NTAG5_STATUS1_ACTIVE_NFC_OK		(1 << 5)
#define NTAG5_STATUS1_GPIO1_IN_STATUS	(1 << 4)
#define NTAG5_STATUS1_GPIO0_IN_STATUS	(1 << 3)
#define NTAG5_STATUS1_ALM_PLM			(1 << 2)
#define NTAG5_STATUS1_I2C_IF_LOCKED		(1 << 1)
#define NTAG5_STATUS1_NFC_IF_LOCKED		(1 << 0)

#define NTAG5_STATUS0_EEPROM_WR_BUSY	(1 << 7)
#define NTAG5_STATUS0_EEPROM_WR_ERROR	(1 << 6)
#define NTAG5_STATUS0_SRAM_DATA_READY	(1 << 5)
#define NTAG5_STATUS0_SYNCH_BLOCK_WRITE	(1 << 4)
#define NTAG5_STATUS0_SYNCH_BLOCK_READ	(1 << 3)
#define NTAG5_STATUS0_PT_TRANSFER_DIR	(1 << 2)
#define NTAG5_STATUS0_VCC_SUPPLY_OK		(1 << 1)
#define NTAG5_STATUS0_NFC_FIELD_OK		(1 << 0)

/* NOTE: 32-bit CONFIG_REG is split into:
 * - 4x 8-bit CONFIG_[3:0] in the datasheet
 * - 2x 16-bit CONFIG[1:0] in KW41 sample code
 * 
 * This file follows the datasheet.
 */

#define NTAG5_CONFIG1_ALM_PLM			(1 << 6)
#define NTAG5_CONFIG1_USE_CASE_CONF(x)	(((x) & 0b11) << 4)
#define NTAG5_CONFIG1_ARBITER_MODE(x)	(((x) & 0b11) << 2)
#define NTAG5_CONFIG1_SRAM_ENABLED		(1 << 1)
#define NTAG5_CONFIG1_PT_TRANSFER_DIR	(1 << 0)

#define NTAG5_CONFIG0_SRAM_COPY_EN		(1 << 7)
#define NTAG5_CONFIG0_DISABLE_NFC		(1 << 5)
#define NTAG5_CONFIG0_AUTO_STANDBY_MODE_EN (1 << 0)

#define NTAG5_CONFIG2_GPIO1_IN(x)		(((x) & 0b11) << 6)
#define NTAG5_CONFIG2_GPIO0_IN(x)		(((x) & 0b11) << 4)
#define NTAG5_CONFIG2_EXTENDED_COMMANDS_SUPPORTED (1 << 3)
#define NTAG5_CONFIG2_LOCK_BLOCK_COMMAND_SUPPORTED (1 << 2)
#define NTAG5_CONFIG2_GPIO1_SLEW_RATE	(1 << 1)
#define NTAG5_CONFIG2_GPIO0_SLEW_RATE	(1 << 0)

#define NTAG5_ED_CONFIG_MASK			(0b1111)
#define NTAG5_ED_CONFIG_DISABLE_ED		(0)
#define NTAG5_ED_CONFIG_NFC_FIELD_DETECT (1)
#define NTAG5_ED_CONFIG_PWM				(2)
#define NTAG5_ED_CONFIG_I2C_TO_NFC		(3)
#define NTAG5_ED_CONFIG_NFC_TO_I2C		(4)
#define NTAG5_ED_CONFIG_ARBITER_LOCK	(5)
#define NTAG5_ED_CONFIG_NDEF_TLV_LENGTH	(6)
#define NTAG5_ED_CONFIG_STANDBY_MODE	(7)
#define NTAG5_ED_CONFIG_WRITE_CMD		(8)
#define NTAG5_ED_CONFIG_READ_CMD		(9)
#define NTAG5_ED_CONFIG_START_OF_CMD	(10)
#define NTAG5_ED_CONFIG_READ_SYNCH_BLOCK (11)
#define NTAG5_ED_CONFIG_WRITE_SYNCH_BLOCK (12)
#define NTAG5_ED_CONFIG_SW_INTERRUPT	(13)

#define NTAG5_I2C_SLAVE_CONF_I2C_WDT_EXPIRED		(1 << 4)
#define NTAG5_I2C_SLAVE_CONF_I2C_SOFT_RESET			(1 << 3)
#define NTAG5_I2C_SLAVE_CONF_I2C_S_REPEATED_START	(1 << 2)
#define NTAG5_I2C_SLAVE_CONF_DISABLE_I2C			(1 << 0)

#define NTAG5_ARBITER_MODE_NORMAL		(0)
#define NTAG5_ARBITER_MODE_MIRROR		(1)
#define NTAG5_ARBITER_MODE_PASSTHROUGH	(2)
#define NTAG5_ARBITER_MODE_PHDC			(3)



#endif /* _NTAG5_H_ */
