/*
Copyright 2021 nxa07283, NXP
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    (1) Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

    (2) Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.  
    
    (3)The name of the author may not be used to
    endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <stdbool.h>
#include <errno.h>
#include <unistd.h>
#include <ntag5.h>
#include <platform/linux/i2c.h>

ntag_t ntag5_open(int type, struct ntag5_hw_config * hw_config)
{
	int res;
	
	if(!hw_config) return NULL;
	if(type != NTAG_TYPE_NTAG5) return NULL;
	
	/* determine the i2c address based on type */
	if(hw_config->i2c_address == NTAG_I2C_ADDR_AUTO) {
		switch(type) {
		case NTAG_TYPE_NTAG5:
			hw_config->i2c_address = NTAG_TYPE_NTAG5;
			break;
		default:
			return NULL;
		};
	};
	
	ntag_t ntag;
	
	ntag = malloc(sizeof(struct ntag));
	if(ntag == NULL) {
		return NULL;
	}
	
	res = i2c_open(ntag, hw_config);
	if(res < 0) {
		return NULL;
	};
	
	return ntag;
}


int ntag5_close(ntag_t ntag)
{
	if(ntag) {
		i2c_close(ntag);
		free(ntag); 
		return 0;
	} else {
		return -EINVAL;	
	}
}


int ntag5_nfc_enable(ntag_t ntag, bool enable)
{
	int res;
	
	res = ntag5_reg_write(ntag, 
		NTAG5_CONFIG_0_REG, 
		NTAG5_CONFIG0_DISABLE_NFC, enable ? 0 : NTAG5_CONFIG0_DISABLE_NFC);
	return res;
}

int ntag5_nfc_is_enabled(ntag_t ntag)
{
	int res;
	uint16_t val;
	
	res = ntag5_reg_read(ntag, NTAG5_CONFIG_0_REG, &val);
	if(res) {
		return res;
	};
	
	return (val & NTAG5_CONFIG0_DISABLE_NFC) ? false : true;
}


int ntag5_nfc_is_present(ntag_t ntag)
{
	int res;
	uint16_t val;
	
	res = ntag5_reg_read(ntag, NTAG5_STATUS0_REG, &val);
	if(res) {
		return res;
	};

	return (val & NTAG5_STATUS0_NFC_FIELD_OK) ? true : false;	
}

int ntag5_reset(ntag_t ntag)
{
	int res;
	uint16_t val;
	
	int cnt;
	
	res = ntag5_reg_write(ntag, NTAG5_RESET_GEN_REG, 0xFF, 0xE7);
	if(res) {
		if(errno != ENXIO) { /* reset initiated during i2c transaction, so no ack */
			return res;
		};
	};
	
	
	cnt = 10;
	do {
		val = 0;
		res = ntag5_reg_read(ntag, NTAG5_STATUS1_REG, &val);
		if(res < 0) {
			if(errno != ENXIO) {
				return res;
			};
		};
		
		if(val & NTAG5_STATUS1_VCC_BOOT_OK) {
			break;
		};
		
		usleep(2 * 1000);	
	} while(--cnt);
	
	if(cnt == 0) { /* timeout */
		errno = ETIME;
		return -1;
	}
	
	//printf("vcc ok\n");
	
	return 0;
};

int ntag5_set_mode(ntag_t ntag, int mode)
{
	int res;
	uint16_t val;

	res = ntag5_reg_write(ntag, NTAG5_CONFIG_1_REG, 0x0C, (mode == NTAG5_MODE_MIRROR) ? (1 << 2) : (0 << 2));
	if(res < 0) {
		return res;
	};
	
	return 0;
};

int ntag5_wait_field(ntag_t ntag, int timeout) {
	
	#define WAIT (4 * 1000)
	int res;
	
	if(timeout == 0) {
		timeout = -1;
	};
	
	while(timeout) {
		res = ntag5_nfc_is_present(ntag);
		if(res) { /* either true or -1 */
			return res;
		};
	
		if(timeout > WAIT) { 
			usleep(WAIT * 1000); 
			timeout -= WAIT;
		} else {
			usleep(timeout * 1000);
			if(timeout > 0) timeout = 0;
		};
		
	};
	
	return false;	
};

int ntag5_pthru_send(ntag_t ntag, uint8_t *buf, int cnt) 
{
	int res;
	uint16_t val;

	res = ntag5_nfc_is_present(ntag);
	if(res != true) {
		errno = ENOLINK;
		return -1;
	};
	
	res = ntag5_reg_write(ntag, NTAG5_CONFIG_1_REG, 0b1101, 0b1000);
	if(res < 0) return res;
	
	res = ntag5_mem_write(ntag, NTAG5_SRAM(NTAG5_SRAM_SIZE - cnt), buf, cnt);
	if(res < 0) return res;
	
	
	do {
		res = ntag5_reg_read(ntag, NTAG5_STATUS_REG, &val);
		if(res < 0) return res;

		if(!(val & NTAG5_STATUS0_NFC_FIELD_OK)) {
			res = ntag5_reg_write(ntag, NTAG5_CONFIG_1_REG, 0b1101, 0b0000);
			if(res < 0) return res;			
			
			errno = ENOLINK;
			return -1;
		};
					
	} while(val & NTAG5_STATUS0_SRAM_DATA_READY);
	
	return 0;
};

int ntag5_pthru_receive(ntag_t ntag, uint8_t *buf, int cnt) 
{
	int res;
	uint16_t val;

	res = ntag5_nfc_is_present(ntag);
	if(res != true) {
		errno = ENOLINK;
		return -1;
	};
	
	res = ntag5_reg_write(ntag, NTAG5_CONFIG_1_REG, 0b1101, 0b1001);
	if(res < 0) return res;
	
	do {
		res = ntag5_reg_read(ntag, NTAG5_STATUS_REG, &val);
		if(res < 0) return res;

		if(!(val & NTAG5_STATUS0_NFC_FIELD_OK)) {
			res = ntag5_reg_write(ntag, NTAG5_CONFIG_1_REG, 0b1101, 0b0000);
			if(res < 0) return res;
			
			errno = ENOLINK;
			return -1;
		};
	} while(!(val & NTAG5_STATUS0_SRAM_DATA_READY));

	/* exit pass-through mode to prevent nfc side from writing more data after 
	 * mem_read clears SRAM_DATA_READY */
	 
	res = ntag5_reg_write(ntag, NTAG5_CONFIG_1_REG, 0b1101, 0b0000);
	if(res < 0) return res;
	
	/* DBG */
	ntag5_reg_read(ntag, NTAG5_STATUS0_REG, &val);
	fprintf(stderr, "s0: %02X\n", val);
	

	res = ntag5_mem_read(ntag, NTAG5_SRAM(NTAG5_SRAM_SIZE - cnt), buf, cnt);
	if(res < 0) return res;

	return 0;
};


