/*
Copyright (c) 2021 nxa07283, NXP
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    (1) Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

    (2) Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.  
    
    (3)The name of the author may not be used to
    endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <error.h>
#include <sys/ioctl.h>
#include <linux/i2c.h>
#include <linux/i2c-dev.h>
#include <platform/linux/i2c.h>
#include <ntag5.h>

#define IS_EEPROM(x) ( ((x) >= NTAG5_EEPROM_START && (x) <= NTAG5_EEPROM_END) ? true : false )
#define IS_SRAM(x) ( ((x) >= NTAG5_SRAM_START && (x) <= NTAG5_SRAM_END) ? true : false )
#define IS_CONFIG(x) ( ((x) >= NTAG5_CONFIG_START && (x) <= NTAG5_CONFIG_END) ? true : false )



int ntag5_reg_read(ntag_t ntag, uint32_t reg, uint16_t *val)
{
    int res;
    int reg_value;
    int width;

    if(ntag == NULL)
    {
		errno = EINVAL;
        return -1;
    }

    uint8_t req[3], resp[1];
    
    if(NTAG5_REG_CHECK(reg)) {
		req[0] = (uint8_t)(NTAG5_REG_GET_ADDR_H(reg));
		req[1] = (uint8_t)(NTAG5_REG_GET_ADDR_L(reg));
		req[2] = (uint8_t)(NTAG5_REG_GET_OFF(reg));
		width = NTAG5_REG_GET_W(reg);
	} else {
		return -1;
	};
    
    struct i2c_rdwr_ioctl_data msgset;
    struct i2c_msg msg[2] = {
		{
			.flags = 0,
			.len = 3,
			.buf = req
			
		},
		{
			.flags = I2C_M_RD,
			.len = 1,
			.buf = &(resp[0])
		}		
	};

	msgset.msgs = msg;
	msgset.nmsgs = 2;
    
   /* Read the Register byte */
    res = i2c_rdwr(ntag, &msgset);
    if(res < 0) {
		return -1;
	};
	
	reg_value = resp[0];
	
	if(width == 2) {
		req[2] += 1;				/* increase reg. address */

		res = i2c_rdwr(ntag, &msgset);
		if(res < 0) {
			return -1;
		};
		
		reg_value |= resp[0] << 8;
	};
	
    *val = reg_value;
    
    return 0;
}


int ntag5_reg_write(ntag_t ntag, uint32_t reg, uint16_t mask, uint16_t val)
{
    int res;
    int width;

    if(ntag == NULL)
    {
		errno = EINVAL;
        return -1;
    }
    
    uint8_t req[5];
    if(NTAG5_REG_CHECK(reg)) {
		req[0] = (uint8_t)(NTAG5_REG_GET_ADDR_H(reg));
		req[1] = (uint8_t)(NTAG5_REG_GET_ADDR_L(reg));
		req[2] = (uint8_t)(NTAG5_REG_GET_OFF(reg));
		width = NTAG5_REG_GET_W(reg);
	} else {
		return -1;
	};

    req[3] = (uint8_t)(mask & 0xff);
    req[4] = (uint8_t)(val & 0xff);

    
    struct i2c_rdwr_ioctl_data msgset;
    struct i2c_msg msg[1] = {
		{
			.flags = 0,
			.len = 5,
			.buf = req
		},
	};

	msgset.msgs = msg;
	msgset.nmsgs = 1;
    
  
  
    /* Read the Register byte */
    
    res = i2c_rdwr(ntag, &msgset);
    if(res < 0) {
		return -1;
	};
	
	if(width == 2) {
		req[2] += 1;
		req[3] = mask >> 8;			/*  */
		req[4] = val >> 8;

		res = i2c_rdwr(ntag, &msgset);
		if(res < 0) {
			return -1;
		};
	};
    
    return 0;
}


int ntag5_block_read(ntag_t ntag, int block, uint8_t *bytes, int len) 
{	
	// fprintf(stderr, "%s: block %04X (%d bytes)\n", __func__, block, len);
	
	int res;

    if(ntag == NULL)
    {
		errno = EINVAL;
        return -1;
    }
    
    if(!IS_SRAM(block) && !IS_EEPROM(block) && !IS_CONFIG(block))
    {
		errno = EFAULT;
		return -1;
	};
	
    uint8_t req[2];
    req[0] = (uint8_t)(block >> 8);
    req[1] = (uint8_t)(block);

    struct i2c_rdwr_ioctl_data msgset;
    struct i2c_msg msg[2] = {
		{
			.flags = 0,
			.len = 2,
			.buf = req
			
		},
		{
			.flags = I2C_M_RD,
			.len = len,
			.buf = bytes
		}		
	};

	msgset.msgs = msg;
	msgset.nmsgs = 2;
	
	res = i2c_rdwr(ntag, &msgset);
	if(res < 0) {
		error_at_line(0, errno, __FILE__, __LINE__, "I2C_RDWR");
		return -1;
	};

	return 0;
}


int ntag5_block_write(ntag_t ntag, int block, uint8_t *bytes, int len) 
{
	// fprintf(stderr, "%s: block %04X (%d bytes)\n", __func__, block, len);
	
	int res;
	bool is_eeprom;
	
	if(IS_EEPROM(block)) {
		is_eeprom = true;
	} else if(IS_SRAM(block) || IS_CONFIG(block)) {
		is_eeprom = false;
	} else {
		errno = EFAULT;
		return -1;
	};

    if(ntag == NULL)
    {
		errno = EINVAL;
        return -1;
    }
    
    if(len > 256 || len % 4)
    {
		errno = EINVAL;
		return -1;
	}
    
	
	/* NOTE: len must be 4 when writing EEPROM, 4*n when writing SRAM */
	
    uint8_t req[2 + 256];
    req[0] = (uint8_t)(block >> 8);
    req[1] = (uint8_t)(block);
    
    for(int i = 0; i < len; i++) {
		req[i+2] = bytes[i];
	};

    struct i2c_rdwr_ioctl_data msgset;
    struct i2c_msg msg[1] = {
		{
			.flags = 0,
			.len = 2 + len,
			.buf = req
		},
	};

	msgset.msgs = msg;
	msgset.nmsgs = 1;
	
	res = i2c_rdwr(ntag, &msgset);
	if(res < 0) {
		return -1;
	};
	
	
	uint16_t status_reg;

	if(is_eeprom) {

		/* wait for completion */
		do {
			usleep(1000);
			res = ntag5_reg_read(ntag, NTAG5_STATUS0_REG, &status_reg);
			if(res) {
				return -1;
			};

		} while (status_reg & NTAG5_STATUS0_EEPROM_WR_BUSY);
		
		if(status_reg & NTAG5_STATUS0_EEPROM_WR_ERROR) {
			/* IO error occurred first, ignore result of ntag5_write_reg */
			ntag5_reg_write(ntag, NTAG5_STATUS0_REG, NTAG5_STATUS0_EEPROM_WR_ERROR, 0);
			errno = -EIO;
			return -1;
		};

	};

	return 0;
}



/* NOTE: this uses linear (byte, not block) addressing */

int ntag5_mem_read(ntag_t ntag, uint16_t addr, uint8_t *bytes, int len)
{	
	int res;
	bool is_eeprom;
	int block = addr / 4;
	int last_block = ((addr + len + 3) / 4) - 1; /* TODO: check, rewrite*/
	int bytes_left = len;
	uint8_t buf[4];

	if(IS_EEPROM(block)) {
		is_eeprom = true;
	} else if(IS_SRAM(block)) {
		is_eeprom = false;
	} else {
		errno = EFAULT;
		return -1;
	};
	
	if(bytes_left == 0) {
		return 0;
	};

	if(is_eeprom) {
		if(!IS_EEPROM(last_block)) {
			errno = EFAULT;
			return -1;
		};
	} else {
		if(!IS_SRAM(last_block)) {
			errno = EFAULT;
			return -1;
		};
	};

	
	/* align to block boundary */
	if(addr % 4) {
		
		res = ntag5_block_read(ntag, block, buf, 4);
		if(res) {
			return -1;
		};
		
		while(addr % 4 && bytes_left) {
			(bytes++)[0] = buf[addr % 4];
			addr += 1;
			bytes_left -= 1;
		};
		block += 1;
		
	};
	
	#define RD_SIZE (64)
	#define MIN(a,b) (((a) < (b)) ? a : b )
	
	while(bytes_left) {
	
		/* from first block boundary */
		res = ntag5_block_read(ntag, block, bytes, MIN(bytes_left, RD_SIZE)); 
		if(res) {
			error_at_line(0, errno, __FILE__, __LINE__, "Mem read error at %04X", addr);
			return -1;
		};
		usleep(10);
		
		addr += RD_SIZE;
		block += RD_SIZE / 4;
		bytes += RD_SIZE;
		bytes_left -= MIN(bytes_left, RD_SIZE);
	
	};
	
	return 0;
}

int ntag5_mem_write(ntag_t ntag, uint16_t addr, uint8_t *bytes, int len)
{
	
	int res;
	bool is_eeprom;
	int block = addr / 4;
	int last_block = ((addr + len + 3) / 4) - 1; /* TODO: check, rewrite*/
	int bytes_left = len;
	uint8_t buf[4];

	if(IS_EEPROM(block)) {
		is_eeprom = true;
	} else if(IS_SRAM(block)) {
		is_eeprom = false;
	} else {
		errno = EINVAL;
		return -1;
	};
	
	if(bytes_left == 0) {
		return 0;
	};

	if(is_eeprom) {
		if(!IS_EEPROM(last_block)) {
			errno = EFAULT;
			return -1;
		};
	} else {
		if(!IS_SRAM(last_block)) {
			errno = EFAULT;
			return -1;
		};
	};

	/* align to block boundary: 
	 * read/modify/write the first block, if not aligned */
	 
	if(addr % 4) {
		
		res = ntag5_block_read(ntag, block, buf, 4);
		if(res) {
			error_at_line(0, errno, __FILE__, __LINE__, "Memory read error at block %04X (first)", block);
			return -1;
		};
		
		while(addr % 4 && bytes_left) {
			buf[addr % 4] = *(bytes++);
			addr += 1;
			bytes_left -= 1;
		};
		
		res = ntag5_block_write(ntag, block, buf, 4);
		if(res) {
			error_at_line(0, errno, __FILE__, __LINE__, "Memory write error at block %04X (first)", block);
			return -1;
		};

		block += 1;
	};

	/* NTAG5 supports single block writes to EEPROM,
	 * and multi-block writes to SRAM */
	
	if(is_eeprom) {
		while(bytes_left >= 4) {
			res = ntag5_block_write(ntag, block, bytes, 4);
			if(res) {
				error_at_line(0, errno, __FILE__, __LINE__, "Memory write error at block %04X", block);
				return -1;
			};

			addr += 4;
			block += 1;
			bytes += 4;
			bytes_left -= 4;
		};
	} else {
		int len_aligned = 4 * (bytes_left / 4);
		res = ntag5_block_write(ntag, block, bytes, len_aligned);
		if(res) {
			error_at_line(0, errno, __FILE__, __LINE__, "Memory write error at block %04X", block);
			return -1;
		};

		addr += len_aligned;
		block += len_aligned / 4;
		bytes_left -= len_aligned;
		bytes += len_aligned;
	
	};
	
	/* read-modify-write the remaining bytes to the last block affected */
	
	if(bytes_left) {
		res = ntag5_block_read(ntag, block, buf, 4);
		if(res) {
			error_at_line(0, errno, __FILE__, __LINE__, "Memory read error at block %04X (last)", block);
			return -1;
		};
				
		while(bytes_left) {
			buf[addr % 4] = *bytes;
			addr += 1;
			bytes += 1;
			bytes_left -= 1;
		};
			
		res = ntag5_block_write(ntag, block, buf, 4);
		if(res) {
			error_at_line(0, errno, __FILE__, __LINE__, "Memory write error at block %04X (last)", block);
			return -1;
		};		
	};
	
	return 0;
}



int ntag5_cfg_write(ntag_t ntag, uint16_t block, uint8_t *bytes)
{	
	int res;
	
	if(!IS_CONFIG(block)) {
		errno = EFAULT;
		return -1;
	};
	
	res = ntag5_block_write(ntag, block, bytes, 4);
	if(res < 0) {
		error_at_line(0, errno, __FILE__, __LINE__, "Config write error at block %04X", block);
		return -1;
	};
	
	return 0;
}


int ntag5_cfg_read(ntag_t ntag, uint16_t block, uint8_t *bytes)
{
	int res;
	
	if(!IS_CONFIG(block)) {
		errno = EFAULT;
		return -1;
	};

	res = ntag5_block_read(ntag, block, bytes, 4);
	if(res < 0) {
		error_at_line(0, errno, __FILE__, __LINE__, "Config read error at block %04X", block);
		return -1;
	};
	
	return 0;
}




