/*
Copyright (c) 2021 nxa07283, NXP
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    (1) Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

    (2) Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.  
    
    (3)The name of the author may not be used to
    endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <linux/i2c-dev.h>
#include <ntag5.h>


#include <argp.h>
#include <stdbool.h>
#include <error.h>
#include <string.h>



struct arguments {
	int i2c_adapter;
	int i2c_address;

	bool write;
	bool sram;
	bool nfc_off;
	bool reset;
};

#define ARG_ADAPTER 		(1000)
#define ARG_ADDRESS 		(1001)

static int parse_opt(int key, char *arg, struct argp_state *state)
{
	//static int arg_count = 0;
	struct arguments *arguments = state->input;
	
	switch(key) {
	case ARG_ADAPTER:
		arguments->i2c_adapter = strtoul(arg, NULL, 10);
		break;
		
	case ARG_ADDRESS:
		arguments->i2c_address = strtoul(arg, NULL, 16);
		break;
		
	case 'w':
		arguments->write = true;
		break;
		
	case 's': 
		arguments->sram = true;
		break;

	case 'n': 
		arguments->nfc_off = true;
		break;
		
	case 'r':
		arguments->reset = true;
		break;
		
	};
	
	return 0;
}


int read_hex(uint8_t *buf, int length) {
	
	int res;
	int cnt = 0;
	
	while(cnt < length) {
		res = fscanf(stdin, "%02hhX", &(buf[cnt]));
		if(res != 1) {
			return cnt;
		};
		
		cnt += 1;
	};
	
	return cnt;
};
	
	



int main(int argc, char *argv[]) 
{
	ntag_t ntag; 
	int res;	
	
	uint8_t payload[NTAG5_EEPROM_SIZE];
	uint8_t mem[NTAG5_EEPROM_SIZE];
	int size;
	int pos;
	
	struct arguments arguments = {
		.i2c_adapter = 0,
		.i2c_address = NTAG_I2C_ADDR_NTAG5,
		.write = false,
		.nfc_off = false,
		.sram = false,
		.reset = false
		
	};

	struct argp_option options[] = {
		{ 0, 0, 0, 0, "I2C options:", 1 },
		{ "adapter", ARG_ADAPTER, "N", 0, "Adapter id (/dev/i2c-N)" },
		{ "address", ARG_ADDRESS, "I2C-ADDR", 0, "Device address (default: 0x54)" },
				
		{ "--write", 'w', 0, 0, "Read hex-encoded NDEF message from stdin and write to tag" },
		{ "--sram", 's', 0, 0, "Write to SRAM (default: EEPROM)" },
		{ "--nfc-off", 'n', 0, 0, "Leave NFC disabled" },
		{ "--reset", 'r', 0, 0, "Reset the tag" },

		{ 0 }
	};
	
	struct argp argp = { options, parse_opt, 0, 0 };
	argp_parse(&argp, argc, argv, 0, 0, &arguments);	/* may exit() when arguments invalid */
	
	struct ntag5_hw_config settings = {
		.i2c_adapter = arguments.i2c_adapter,
		.i2c_address = arguments.i2c_address
	};
	
	
	ntag = ntag5_open(NTAG_TYPE_NTAG5, &settings);
	if(ntag == NULL) {
		fprintf(stderr, "Error accessing the tag\n");
		return 0;
	};
	
	if(arguments.reset) {
		res = ntag5_reset(ntag);
		if(res < 0) return -1;
	};
	
	res = ntag5_nfc_enable(ntag, false);
	if(res < 0) {
		fprintf(stderr, "Could not disable nfc\n");
		return -1;
	};
	
	if(arguments.write) {
		if(arguments.sram) {
			size = NTAG5_SRAM_SIZE - 16;
		} else {
			size = NTAG5_EEPROM_SIZE - 32;
		};
		
		/* read up to size bytes from stdin */
		printf("%d, %d\n", size, arguments.sram);
		size = read_hex(payload, size);
		printf("%d\n", size);

		/* Capability Container */
		mem[0] = 0xE1;
		mem[1] = 0x40;
		mem[2] = arguments.sram ? 0x08 : 0x80;
		mem[3] = 0x09;
		
		/* NDEF message TLV */
		mem[4] = 0x03;
		
		if(size < 255) {
			mem[5] = size;
			pos = 6;
		} else {
			mem[5] = 0xFF;
			mem[6] = (size >> 8) & 0xFF;
			mem[7] = size & 0xFF;
			pos = 8;
		};

		memcpy(mem + pos, payload, size);
		pos += size;
		
		/* pad with NULL TLVs to block boundary */
		while(pos % 4) {
			mem[pos] = 0;
			pos += 1;
		};
		
		/* Terminator TLV + padding */
		mem[pos++] = 0xFE;
		mem[pos++] = 0x00;
		mem[pos++] = 0x00;
		mem[pos++] = 0x00;
		
		size = pos;
		

		if(arguments.sram) {
			res = ntag5_mem_write(ntag, NTAG5_SRAM(0), mem, size);
		} else {
			res = ntag5_mem_write(ntag, NTAG5_EEPROM(0), mem, size);
		};
		
		if(res < 0) {
			fprintf(stderr, "Write error\n");
			return -1;
		};

	};
	
	if(arguments.sram) {
		res = ntag5_set_mode(ntag, NTAG5_MODE_MIRROR);
	} else {
		res = ntag5_set_mode(ntag, NTAG5_MODE_NORMAL);
	};
	
	if(res < 0) {
		fprintf(stderr, "Could not set mode\n");
		return -1;
	};
	
	if(!arguments.nfc_off) {
		res = ntag5_nfc_enable(ntag, true);
	};
	
	if(res < 0) {
		fprintf(stderr, "Could not enable nfc\n");
		return -1;
	};

	
	ntag5_close(ntag);
	
	return 0;
}
