/*
Copyright (c) 2021 nxa07283, NXP
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    (1) Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

    (2) Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.  
    
    (3)The name of the author may not be used to
    endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <linux/i2c-dev.h>
#include <ntag5.h>


#include <argp.h>
#include <stdbool.h>
#include <error.h>
#include <string.h>
#include <unistd.h>



struct arguments {
	int i2c_adapter;
	int i2c_address;

	int op;
		

	bool verbose;
	bool hex;

};

#define ARG_ADAPTER 		(1000)
#define ARG_ADDRESS 		(1001)
#define ARG_WRITE_CONFIG	(1002)
#define ARG_READ_CONFIG		(1003)
#define ARG_YES				(1004)




static int parse_opt(int key, char *arg, struct argp_state *state)
{
	//static int arg_count = 0;
	struct arguments *arguments = state->input;
	
	switch(key) {
	case ARG_ADAPTER:
		arguments->i2c_adapter = strtoul(arg, NULL, 10);
		break;
		
	case ARG_ADDRESS:
		arguments->i2c_address = strtoul(arg, NULL, 16);
		break;
		
	case 'v': 
		arguments->verbose = true;
		break;

	case 'x': 
		arguments->hex = true;
		break;

	};
	
	return 0;
}


int read_hex(uint8_t *buf, int length) {
	
	int res;
	int cnt = 0;
	
	while(cnt < length) {
		res = fscanf(stdin, "%02hhX", &(buf[cnt]));
		if(res != 1) {
			return cnt;
		};
		
		cnt += 1;
	};
	
	return cnt;
};
	
	

int main(int argc, char *argv[]) 
{
	ntag_t ntag; 
	
	
	
	uint8_t sram_data[256];
	int res;
	uint16_t val;
	int cnt;
	
	struct arguments arguments = {
		.i2c_adapter = 0,
		.i2c_address = NTAG_I2C_ADDR_NTAG5,
		.verbose = false,
		.hex = false,
	};

	struct argp_option options[] = {
		{ 0, 0, 0, 0, "I2C options:", 1 },
		{ "adapter", ARG_ADAPTER, "N", 0, "Adapter id (/dev/i2c-N)" },
		{ "address", ARG_ADDRESS, "I2C-ADDR", 0, "Device address (default: 0x54)" },
				
		{ 0, 0, 0, 0, "Other:", 7 },
		{ "hex", 'x', 0, 0, "Output ascii-encoded hex data" },
		{ "verbose", 'v', 0, 0, "Verbose output" },
		{ 0 }
	};
	
	struct argp argp = { options, parse_opt, 0, 0 };
	argp_parse(&argp, argc, argv, 0, 0, &arguments);	/* may exit() when arguments invalid */
	
	struct ntag5_hw_config settings = {
		.i2c_adapter = arguments.i2c_adapter,
		.i2c_address = arguments.i2c_address
	};
	
	
	if(arguments.verbose) {	
			fprintf(stderr, "Opening /dev/i2c-%d:0x%02X\n", 
				arguments.i2c_adapter, arguments.i2c_address);
	};
	
	
	ntag = ntag5_open(NTAG_TYPE_NTAG5, &settings);
	if(ntag == NULL) {
		fprintf(stderr, "Error accessing the tag\n");
		return 0;
	};
	
	errno = 0;
	res = ntag5_reset(ntag);
	printf("reset: %s\n", (res == 0) ? "ok" : "failed");
	if(res < 0) return -1;
	
	
	while(1) {
		cnt = 0;

		fprintf(stderr, "waiting for NFC...");
		do {
			res = ntag5_nfc_is_present(ntag);
			if(res < 0) return -1;
			usleep(10 * 1000);
		} while(!res);
		
		fprintf(stderr, "\n");

		usleep(100 * 1000);

		int dir = 1;

		while(true) {
			
			if(dir) { /* i2c -> nfc */

				memset(sram_data, 0, 256);
				sram_data[0] = cnt;
				fprintf(stderr, "send (0x%02X)\n", cnt);
				cnt += 1;
				
				res = ntag5_pthru_send(ntag, sram_data, 256);
				if(res < 0) {
					if(errno == ENOLINK) break;
					else return res;
				};

			} else { /* nfc -> i2c */

				fprintf(stderr, "receive\n");
				res = ntag5_pthru_receive(ntag, sram_data, 240);
				if(res < 0) {
					if(errno == ENOLINK) break;
					else return res;
				};
				
				printf("\n");
				for(int i = 0; i < 240; i++) {
					printf("%02x ", sram_data[i]);
				};
					
				printf("\n");
					
			};
			
			dir ^= 1;

		};

	};

	ntag5_close(ntag);
	
	return 0;
}
