/*
Copyright 2021 nxa07283, NXP
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

    (1) Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer. 

    (2) Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.  
    
    (3)The name of the author may not be used to
    endorse or promote products derived from this software without
    specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING
IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE. 
*/

#include <stdio.h>
#include <stdlib.h>
#include <linux/i2c-dev.h>
#include <ntag5.h>
#include <argp.h>
#include <stdbool.h>
#include <error.h>

#define OP_NOOP (0)
#define OP_READ_REG (1)
#define OP_WRITE_REG (2)
#define OP_READ_MEM (3)
#define OP_WRITE_MEM (4)
#define OP_READ_CONFIG (5)
#define OP_WRITE_CONFIG (6)

struct arguments {
	int i2c_adapter;
	int i2c_address;

	int op;
		
	/* reg access */
	uint32_t reg;
	int offset;
	int mask;
	int bytes;
	uint32_t value;
	bool sram;
	
	/* mem access */
	uint32_t addr;
	int length;
	
	bool yes;
	bool verbose;
	bool hex;

};

#define ARG_ADAPTER 		(1000)
#define ARG_ADDRESS 		(1001)
#define ARG_WRITE_CONFIG	(1002)
#define ARG_READ_CONFIG		(1003)
#define ARG_YES				(1004)


static int parse_opt(int key, char *arg, struct argp_state *state)
{
	static int arg_count = 0;
	struct arguments *arguments = state->input;
	
	switch(key) {
	case ARG_ADAPTER:
		arguments->i2c_adapter = strtoul(arg, NULL, 10);
		break;
		
	case ARG_ADDRESS:
		arguments->i2c_address = strtoul(arg, NULL, 16);
		break;
		
	case 'v': 
		arguments->verbose = true;
		break;

	case 'x': 
		arguments->hex = true;
		break;
		
	case 'r':
		arguments->op = OP_READ_MEM;
		arguments->addr = strtoul(arg, NULL, 16);
		arg_count = 1;
		break;

	case 'w':
		arguments->op = OP_WRITE_MEM;
		arguments->addr = strtoul(arg, NULL, 16);
		arg_count = 1;
		break;

	case 's':
		arguments->sram = true;
		break;
		
	case 'e':
		arguments->sram = false;
		break;
	
	case 'R':
		arguments->op = OP_READ_REG;
		arguments->reg = strtoul(arg, NULL, 16);
		break;
		
	case 'W':
		arguments->op = OP_WRITE_REG;
		arguments->reg = strtoul(arg, NULL, 16);
		arg_count = 1;
		break;
		
	case 'b':
		switch(strtoul(arg, NULL, 10)) {
		case 1:
			arguments->bytes = 1;
			break;
		case 2:
			arguments->bytes = 2;
			break;
		default:
			argp_failure(state, 1, 0, "Only 8-bit and 16-bit access is suported");
		};
		break;
		
	case 'm':
		arguments->mask = strtoul(arg, NULL, 16) & 0xFFFF;
		break;

	case 'o':
		arguments->offset = strtoul(arg, NULL, 10);
		if(arguments->offset < 0 || arguments->offset > 3) {
			argp_failure(state, 1, 0, "OFFSET must be between 0 and 3");
		};
		break;

	case ARG_READ_CONFIG:
		arguments->op = OP_READ_CONFIG;
		arguments->reg = strtoul(arg, NULL, 16);
		break;
		
	case ARG_WRITE_CONFIG:
		arguments->op = OP_WRITE_CONFIG;
		arguments->reg = strtoul(arg, NULL, 16);
		arg_count = 1;
		break;
		
	case ARG_YES:
		arguments->yes = true;
		break;

	case ARGP_KEY_ARG:

		if(arg_count == 1) {
			switch(arguments->op) {
			case OP_WRITE_REG: /* fall through */
			case OP_WRITE_CONFIG:
				arguments->value = strtoul(arg, NULL, 16);
				break;
			
			case OP_WRITE_MEM: /* fall through */
			case OP_READ_MEM:
				arguments->length = strtoul(arg, NULL, 0);
				break;
			};
		};
		
		arg_count -= 1;
		break;
		
	case ARGP_KEY_END:
		if(arg_count < 0) {
			argp_failure(state, 1, 0, "Too many arguments");
		};

		if(arg_count > 0) {	
			argp_failure(state, 1, 0, "Expected argument");
		};

		
		break;
	};
	
	return 0;
}



int read_hex(uint8_t *buf, int length) {
	
	int res;
	int cnt = 0;
	
	while(cnt < length) {
		res = fscanf(stdin, "%02hhX", &(buf[cnt]));
		if(res != 1) {
			return cnt;
		};
		
		cnt += 1;
	};
	
	return cnt;
};
	
	



int main(int argc, char *argv[]) 
{
	ntag_t ntag; 
	
	int res;
	uint16_t val;
	uint8_t buf[NTAG5_EEPROM_SIZE];
	
	struct arguments arguments = {
		.i2c_adapter = 0,
		.i2c_address = NTAG_I2C_ADDR_NTAG5,
		.bytes = 1,
		.offset = 0,
		.mask = 0xFFFF,
		.sram = false,
		.verbose = false,
		.hex = false,
		.yes = false,
		.op = OP_NOOP,
	};

	struct argp_option options[] = {
		{ 0, 0, 0, 0, "I2C options:", 1 },
		{ "adapter", ARG_ADAPTER, "N", 0, "Adapter id (/dev/i2c-N)" },
		{ "address", ARG_ADDRESS, "I2C-ADDR", 0, "Device address (default: 0x54)" },
		
		{ 0, 0, 0, 0, "Register access:", 2 },
		{ 0, 'W', "REG VALUE", 0, "Write register" },
		{ 0, 'R', "REG", 0, "Read register" },
		{ 0, 0, 0, 0, 0, 3 },	/* separator to prevent sorting by key */
		{ 0, 'm', "MASK", 0, "Write mask (default: 0xFFFF)" },
		{ 0, 'o', "OFFSET", 0, "Byte offset within page (default: 0)" },
		{ 0, 'b', "BYTES", 0, "Register width, 1 or 2 bytes (default: 1)" },
		
		{ 0, 0, 0, 0, "Register access:", 4 },
		{ "read-config", ARG_READ_CONFIG, "REG", 0, "Read configuration block" },
		{ "write-config", ARG_WRITE_CONFIG, "REG VALUE", 0, "Write configuration block" },
		{ "yes", ARG_YES, 0, 0, "Confirm write config" },
		
		
		{ 0, 0, 0, 0, "Memory access:", 5 },
		{ 0, 'w', "ADDR LENGTH", 0, "Write memory" },
		{ 0, 'r', "ADDR LENGTH", 0, "Read memory" },
		{ 0, 0, 0, 0, 0, 6 },	/* separator to prevent sorting by key */
		{ "sram", 's', 0, 0, "Access SRAM" },
		{ "eeprom", 'e', 0, 0, "Access EEPROM (default)" },
		
		{ 0, 0, 0, 0, "Other:", 7 },
		{ "hex", 'x', 0, 0, "Output ascii-encoded hex data" },
		{ "verbose", 'v', 0, 0, "Verbose output" },
		{ 0 }
	};
	
	struct argp argp = { options, parse_opt, 0, 0 };
	argp_parse(&argp, argc, argv, 0, 0, &arguments);	/* may exit() when arguments invalid */
	
	struct ntag5_hw_config hw_config = {
		.i2c_adapter = arguments.i2c_adapter,
		.i2c_address = arguments.i2c_address
	};
	
	
	if(arguments.verbose) {	
			fprintf(stderr, "Opening /dev/i2c-%d:0x%02X\n", 
				arguments.i2c_adapter, arguments.i2c_address);
	};
	
	ntag = ntag5_open(NTAG_TYPE_NTAG5, &hw_config);
	if(ntag == NULL) {
		fprintf(stderr, "Error accessing the tag\n");
		return 0;
	};
	
	
	//res  = NTAG_WriteRegister(ntag, NTAG_MEM_OFFSET_WDT_EN_REG, 0x1, 0x0);
	//res  = NTAG_WriteRegister(ntag, NTAG_MEM_OFFSET_CONFIG0_REG, 0b00000010, 0b00000010);
	//res |= NTAG_WriteRegister(ntag, NTAG_MEM_OFFSET_CONFIG0_REG, 0b00000001, 0b00000000);
	//if(res < 0) {
	//	fprintf(stderr, "Error setting up\n");
	//	return 0;
	//}	

	switch(arguments.op) {
	case OP_NOOP:
		if(arguments.verbose) {	
			fprintf(stderr, "No operation specified\n");
		};
		break;
		
	case OP_READ_REG:
		if(arguments.verbose) {	
			fprintf(stderr, "Read reg. 0x%04X+%d (%d bits)\n", 
				arguments.reg, 
				arguments.offset, 
				arguments.bytes * 8
			);
		};
		
		
		res = ntag5_reg_read(ntag, 
			NTAG5_REG(
				arguments.reg, 
				arguments.offset, 
				arguments.bytes
			), 
			&val);
			
		if(res < 0) {
			error_at_line(res, errno, __FILE__, __LINE__, "Register read error");
			break;
		};
		
		buf[0] = val & 0xFF;
		buf[1] = (val >> 8) & 0xFF;

		if(arguments.hex) {
			for(int i = arguments.bytes; i; i--) { /* MSB first */
				fprintf(stdout, "%02X", buf[i - 1]);
			};
			fprintf(stdout, "\n");
		} else {
			fwrite(buf, arguments.bytes, 1, stdout); 
		};
		
		break;
		
		
	case OP_WRITE_REG:
		if(arguments.bytes == 1) {
			arguments.mask &= 0xFF;
			arguments.value &= 0xFF;
		};
	
		if(arguments.verbose) {	
			fprintf(stderr, "Write reg. 0x%04X +%d (%d bits) = 0x%04X (mask 0x%04X)\n", 
				arguments.reg, 
				arguments.offset, 
				arguments.bytes * 8,
				arguments.value,
				arguments.mask
			);
		};
		
		res = ntag5_reg_write(ntag, 
			NTAG5_REG(
				arguments.reg, 
				arguments.offset, 
				arguments.bytes
			), 
			arguments.mask, 
			arguments.value
		);

		if(res < 0) {
			error_at_line(res, errno, __FILE__, __LINE__, "Register write error");
		};
		
		break;


	case OP_READ_MEM:
	
		if(arguments.verbose) {	
			fprintf(stderr, "Read %s 0x%04X (%d bytes)\n", 
				arguments.sram ? "SRAM" : "EEPROM",
				arguments.addr, 
				arguments.length
			);
		};

		if(arguments.sram) {
			arguments.addr = NTAG5_SRAM(arguments.addr);
		} else {
			arguments.addr = NTAG5_EEPROM(arguments.addr);
		};
		
		res = ntag5_mem_read(ntag, arguments.addr, buf, arguments.length);
		if(res < 0) {
			error_at_line(0, errno, __FILE__, __LINE__, "Memory read error");
			break;
		};
		
		
		if(arguments.hex) {
			for(int i = 0; i < arguments.length; i++) {
				fprintf(stdout, "%02X", buf[i]);
			};
			fprintf(stdout, "\n");
		} else {
			fwrite(buf, 1, arguments.length, stdout);
		};
		
		break;
		
	case OP_WRITE_MEM:
	
		if(arguments.verbose) {	
			fprintf(stderr, "Write %s 0x%04X (%d bytes)\n", 
				arguments.sram ? "SRAM" : "EEPROM",
				arguments.addr, 
				arguments.length
			);
		};
		
		if(arguments.sram) {
			arguments.addr = NTAG5_SRAM(arguments.addr);
		} else {
			arguments.addr = NTAG5_EEPROM(arguments.addr);
		};	

		if(arguments.hex) {
			res = read_hex(buf, arguments.length);
		} else {
			res = fread(buf, 1, arguments.length, stdin);
		};
		
		if(res != arguments.length) {
			fprintf(stderr, "Error: Expected %d bytes, read %d\n", arguments.length, res);
			break;
		};
	
		res = ntag5_mem_write(ntag, arguments.addr, buf, arguments.length);
		break;


	case OP_READ_CONFIG:
	
	
		if(arguments.verbose) {	
			fprintf(stderr, "Read config page 0x%04X\n", 
				arguments.reg
			);
		};

		res = ntag5_cfg_read(ntag, arguments.reg, buf);
		if(res < 0) {
			error_at_line(0, errno, __FILE__, __LINE__, "Config read error");
			break;
		};

		if(arguments.hex) {
			for(int i = 0; i < 4; i++) {
				fprintf(stdout, "%02X", buf[i]);
			};
			fprintf(stdout, "\n");
		} else {
			fwrite(buf, 1, 4, stdout);
		};
		break;

	case OP_WRITE_CONFIG:
		buf[0] = (arguments.value >> 24);
		buf[1] = (arguments.value >> 16);
		buf[2] = (arguments.value >>  8);
		buf[3] = (arguments.value      );
	
		if(!arguments.yes) {
			fprintf(stderr, "SOME LOCK BITS IN CONFIGURATION AREA ARE WRITE-ONCE!\n"
				"Write %02X%02X%02X%02X to config page 0x%04X? (y/n) ", 
				buf[0], buf[1], buf[2], buf[3],
				arguments.reg
			);

			if(tolower(fgetc(stdin)) != 'y') break;
		};

		ntag5_cfg_write(ntag, arguments.reg, buf);
	
		break;
		
	};


	/*
	res = NTAG_ReadRegister(ntag, NTAG_MEM_OFFSET_TAG_CONFIG0_REG, &val);
	fprintf(stderr, "(%d) NTAG_MEM_OFFSET_TAG_CONFIG0_REG: %04x\n", res, val);

	res = NTAG_WriteRegister(ntag, NTAG_MEM_OFFSET_TAG_CONFIG0_REG, 0x0020, 0x0020);
	fprintf(stderr, "(%d)\n", res);
	
	res = NTAG_ReadRegister(ntag, NTAG_MEM_OFFSET_TAG_CONFIG0_REG, &val);
	fprintf(stderr, "(%d) NTAG_MEM_OFFSET_TAG_CONFIG0_REG: %04x\n", res, val);
	
	fprintf(stderr, "NFC OFF\n");
	fgetc(stdin);
	fprintf(stderr, "NFC ON\n");

	res = NTAG_WriteRegister(ntag, NTAG_MEM_OFFSET_TAG_CONFIG0_REG, 0x0020, 0x0000);
	fprintf(stderr, "(%d)\n", res);
	
	res = NTAG_ReadRegister(ntag, NTAG_MEM_OFFSET_TAG_CONFIG0_REG, &val);
	fprintf(stderr, "(%d) NTAG_MEM_OFFSET_TAG_CONFIG0_REG: %04x\n", res, val);
	*/
	
	
	//res = NTAG_ReadRegister(ntag, NTAG_MEM_OFFSET_TAG_CONFIG0_REG, 0x0020, 0x0020);
	

	/* read all blocks */

	/*
	uint8_t buf[4] = { 0 };
	FILE *fp;
	fp = fopen("eeprom.bin", "w");
	
	for(int i = 0x000; i <= 0x1FF; i += 1) {
		res = NTAG_ReadBlock(ntag, i, buf, 4);
		if(res) {
			fprintf(stderr, "Read error at block %d\n", i);
			break;
		};
		fwrite(buf, 4, 1, fp);
	};
		
	fclose(fp);		
	*/

	
/*
	uint8_t ndefmsg[] = {  
		0xe1, 0x40, 0x80, 0x09,
		0x03, 0x15, 0xd1, 0x01, 
		0x11, 0x54, 0x82, 0x65, 
		0x6e, 0xff, 0xfe, 0x48, 
		0x00, 0x65, 0x00, 0x6c, 
		0x00, 0x6c, 0x00, 0x6f, 
		0x00, 0x21, 0x00, 0xfe, 
		0x00, 0x00, 0x00, 0x00
	};
*/	/*
	uint8_t readback[256];

	int bytes_left = 8 * 4;
	int block = 0;
	uint8_t *pos = ndefmsg;
	*/
	
	
	/*
	while(bytes_left > 0) {
		fprintf(stderr, "Writing block %d: %02X %02X %02X %02X\n", block, pos[0], pos[1], pos[2], pos[3]); 
		NTAG_WriteBlock(ntag, block, pos, 4);
		if(res) {
			fprintf(stderr, "Write error at block %d\n", block);
			break;
		};
		block += 1;
		pos += 4;
		bytes_left -= 4;
	};
	*/
	/*
	NTAG_Write(ntag, 0, ndefmsg, 8*4);

	ndefmsg[16] = '\0';
	ndefmsg[17] = 'y';
	ndefmsg[18] = '\0';
	ndefmsg[19] = 'p';
	ndefmsg[20] = '\0';
	ndefmsg[21] = 'e';
	

	NTAG_Write(ntag, 16, &(ndefmsg[16]), 6);


	res = NTAG_Read(ntag, 0, readback, 256);
	fprintf(stderr, "(%d)\n", res);

	
	FILE *fp;
	fp = fopen("dump.bin", "w");
	fwrite(readback, 256, 1, fp);
	fclose(fp);	
	*/
	/*
	NTAG_WriteRegister(ntag, NTAG_MEM_OFFSET_TAG_CONFIG0_REG, NTAG5_CONFIG0_SRAM_ENABLE, NTAG5_CONFIG0_SRAM_ENABLE);
	
	NTAG_Write(ntag, 0, ndefmsg, 8*4);
	
	
	
	NTAG_WriteRegister(ntag, NTAG_MEM_OFFSET_TAG_CONFIG0_REG, NTAG5_CONFIG0_ARBITER_MODE(0xFF), NTAG5_CONFIG0_ARBITER_MODE(0));
	
	*/
	

	ntag5_close(ntag);
	//printf("\n");
	
	
	return 0;
}
